use std::num::NonZero;

use crate::string_table::{self, ImmOrIndex};

pub type Node = NonZero<u32>;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Tag {
    Undefined,
    Root,
    FunctionProto {
        /// Ident
        name: Node,
        /// ParameterList
        parameters: Option<Node>,
        return_type: Node,
    },
    ParameterList {
        /// [Parameter]
        parameters: Vec<Node>,
    },
    Parameter {
        /// Ident
        name: Node,
        /// TypeName
        ty: Node,
    },
    // TypeName: meta-tag of Ident | PrimitiveType | Pointer
    Pointer {
        /// TypeName
        pointee: Node,
    },
    PointerQualifier {
        constness: bool,
    },
    IntegralType(IntegralType),
    PrimitiveType(PrimitiveType),
    FunctionDecl {
        /// FunctionProto
        proto: Node,
        /// Block
        body: Node,
    },
    Ident {
        name: string_table::Index,
    },
    Constant {
        bytes: ImmOrIndex,
        ty: Type,
    },
    Block {
        /// ReturnStmt | ExprStmt | VarDecl
        statements: Vec<Node>,
        trailing_expr: Option<Node>,
    },
    ReturnStmt {
        expr: Option<Node>,
    },
    ExprStmt {
        expr: Node,
    },
    VarDecl {
        let_or_var: LetOrVar,
        /// Ident
        name: Node,
        /// TypeName
        explicit_type: Option<Node>,
        assignment: Option<Node>,
    },
    GlobalDecl {
        /// Ident
        name: Node,
        /// TypeName
        explicit_type: Option<Node>,
        assignment: Node,
    },
    DeclRef(Node),
    GlobalRef(Node),
    CallExpr {
        /// Ident | Expr
        lhs: Node,
        /// ArgumentList
        rhs: Option<Node>,
    },
    ArgumentList {
        /// [Argument]
        parameters: Vec<Node>,
    },
    Argument {
        /// Ident
        name: Option<Node>,
        /// expr
        expr: Node,
    },
    ExplicitCast {
        lhs: Node,
        /// TypeName
        typename: Node,
    },
    Deref {
        lhs: Node,
    },
    Ref {
        lhs: Node,
    },
    Not {
        lhs: Node,
    },
    Negate {
        lhs: Node,
    },
    Or {
        lhs: Node,
        rhs: Node,
    },
    And {
        lhs: Node,
        rhs: Node,
    },
    BitOr {
        lhs: Node,
        rhs: Node,
    },
    BitAnd {
        lhs: Node,
        rhs: Node,
    },
    BitXOr {
        lhs: Node,
        rhs: Node,
    },
    Eq {
        lhs: Node,
        rhs: Node,
    },
    NEq {
        lhs: Node,
        rhs: Node,
    },
    Lt {
        lhs: Node,
        rhs: Node,
    },
    Gt {
        lhs: Node,
        rhs: Node,
    },
    Le {
        lhs: Node,
        rhs: Node,
    },
    Ge {
        lhs: Node,
        rhs: Node,
    },
    Shl {
        lhs: Node,
        rhs: Node,
    },
    Shr {
        lhs: Node,
        rhs: Node,
    },
    Add {
        lhs: Node,
        rhs: Node,
    },
    Sub {
        lhs: Node,
        rhs: Node,
    },
    Mul {
        lhs: Node,
        rhs: Node,
    },
    Rem {
        lhs: Node,
        rhs: Node,
    },
    Div {
        lhs: Node,
        rhs: Node,
    },
    Assign {
        lhs: Node,
        rhs: Node,
    },
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum LetOrVar {
    Let,
    Var,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub struct IntegralType {
    pub signed: bool,
    pub bits: u16,
}

impl core::fmt::Display for IntegralType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}{}", if self.signed { "i" } else { "u" }, self.bits)
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum FloatingType {
    Binary32,
    Binary64,
}

impl core::fmt::Display for FloatingType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            FloatingType::Binary32 => "binary32",
            FloatingType::Binary64 => "binary64",
        })
    }
}

impl IntegralType {
    pub fn u32() -> IntegralType {
        Self {
            signed: false,
            bits: 32,
        }
    }
    pub fn u64_bitmask(self) -> u64 {
        (1u64 << self.bits) - 1
    }
    pub fn u128_bitmask(self) -> u128 {
        (1u128 << self.bits) - 1
    }
}

#[derive(Debug, Clone, Eq, Hash)]
pub enum Type {
    Any,
    Void,
    Bool,
    ComptimeNumber,
    Integer(IntegralType),
    Floating(FloatingType),
    Pointer {
        constness: bool,
        pointee: Box<Self>,
    },
    Fn {
        parameter_types: Vec<Self>,
        return_type: Box<Self>,
    },
}

impl core::fmt::Display for Type {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Type::Any => f.write_str("?"),
            Type::Void => f.write_str("void"),
            Type::Bool => f.write_str("bool"),
            Type::ComptimeNumber => f.write_str("comptime_number"),
            Type::Integer(t) => t.fmt(f),
            Type::Floating(t) => t.fmt(f),
            Type::Pointer { constness, pointee } => {
                write!(f, "*{}{}", if *constness { "const " } else { "" }, pointee)
            }
            Type::Fn {
                parameter_types,
                return_type,
            } => {
                write!(f, "fn (")?;
                for param in parameter_types.iter().map(|p| Some(p)).intersperse(None) {
                    match param {
                        Some(param) => param.fmt(f)?,
                        None => write!(f, ", ")?,
                    }
                }

                write!(f, ") -> {}", return_type)
            }
        }
    }
}

impl PartialEq for Type {
    fn eq(&self, other: &Self) -> bool {
        match (self, other) {
            (Self::ComptimeNumber, Self::Integer(_)) => true,
            (Self::Integer(_), Self::ComptimeNumber) => true,
            (Self::ComptimeNumber, Self::Floating(_)) => true,
            (Self::Floating(_), Self::ComptimeNumber) => true,
            (Self::Integer(l0), Self::Integer(r0)) => l0 == r0,
            (Self::Floating(l0), Self::Floating(r0)) => l0 == r0,
            (
                Self::Pointer {
                    pointee: l_pointee, ..
                },
                Self::Pointer {
                    pointee: r_pointee, ..
                },
            ) => l_pointee == r_pointee,
            (
                Self::Fn {
                    parameter_types: l_parameter_types,
                    return_type: l_return_type,
                },
                Self::Fn {
                    parameter_types: r_parameter_types,
                    return_type: r_return_type,
                },
            ) => l_parameter_types == r_parameter_types && l_return_type == r_return_type,
            _ => core::mem::discriminant(self) == core::mem::discriminant(other),
        }
    }
}

impl Type {
    pub fn as_primitive_type(&self) -> Option<PrimitiveType> {
        match self {
            Type::Void => Some(PrimitiveType::Void),
            Type::Bool => Some(PrimitiveType::Bool),
            Type::Integer(t) => Some(PrimitiveType::IntegralType(*t)),
            Type::Floating(t) => Some(PrimitiveType::FloatingType(*t)),
            _ => None,
        }
    }

    pub fn pointee(&self) -> Option<&Type> {
        match self {
            Self::Pointer { pointee, .. } => Some(&pointee),
            _ => None,
        }
    }

    pub fn equal_type(&self, rhs: &Self) -> Option<Type> {
        match (self, rhs) {
            (Self::ComptimeNumber, Self::Floating(_))
            | (Self::ComptimeNumber, Self::Integer(_)) => Some(rhs.clone()),
            (Self::Integer(_), Self::ComptimeNumber)
            | (Self::Floating(_), Self::ComptimeNumber) => Some(self.clone()),
            (
                Self::Pointer {
                    constness: a_const,
                    pointee: a_ptr,
                },
                Self::Pointer {
                    constness: b_const,
                    pointee: b_ptr,
                },
            ) => {
                if a_ptr == b_ptr {
                    Some(Self::Pointer {
                        constness: *a_const || *b_const,
                        pointee: a_ptr.clone(),
                    })
                } else {
                    None
                }
            }
            _ => {
                if self.eq(rhs) {
                    Some(self.clone())
                } else {
                    None
                }
            }
        }
    }
    pub fn void() -> Type {
        Self::Void
    }
    pub fn bool() -> Type {
        Self::Void
    }
    pub fn comptime_number() -> Type {
        Self::ComptimeNumber
    }
    pub fn any() -> Type {
        Self::Any
    }
    pub fn into_ptr(self) -> Type {
        Self::Pointer {
            constness: false,
            pointee: Box::new(self),
        }
    }

    pub fn bit_width(&self) -> u16 {
        match self {
            Type::Any => 0,
            Type::Void => 0,
            Type::Bool => 1,
            Type::ComptimeNumber => u16::MAX,
            Type::Integer(i) => i.bits,
            Type::Floating(f) => match f {
                FloatingType::Binary32 => 32,
                FloatingType::Binary64 => 64,
            },
            Type::Pointer { .. } => 64,
            Type::Fn { .. } => 64,
        }
    }

    pub fn can_logical_and_or(&self) -> bool {
        match self {
            Type::Bool => true,
            _ => false,
        }
    }

    pub fn can_negate(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Bool | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn can_bitxor_and_or(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Bool | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn can_add_sub(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Pointer { .. } | Type::Floating(_) | Type::Integer(_) => {
                true
            }
            _ => false,
        }
    }
    pub fn can_shift(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn can_eq(&self) -> bool {
        match self {
            Type::ComptimeNumber
            | Type::Pointer { .. }
            | Type::Bool
            | Type::Floating(_)
            | Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn can_cmp(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Pointer { .. } | Type::Floating(_) | Type::Integer(_) => {
                true
            }
            _ => false,
        }
    }
    pub fn can_mul_div_rem(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Floating(_) | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn is_integer(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn is_float(&self) -> bool {
        match self {
            Type::ComptimeNumber | Type::Floating(_) => true,
            _ => false,
        }
    }
    pub fn remove_ptr(self) -> crate::parser::Result<Self> {
        match self {
            Type::Pointer { pointee, .. } => Ok(Box::into_inner(pointee)),
            _ => Err(crate::parser::Error::TriedToDerefNonPointer),
        }
    }
    pub fn align_of(&self) -> u32 {
        self.size_of()
    }
    pub fn size_of(&self) -> u32 {
        match self {
            Type::Any => 0,
            Type::Void => 0,
            Type::ComptimeNumber => 0,
            Type::Bool => 1,
            Type::Integer(t) => t.bits.div_ceil(8) as u32,
            Type::Floating(t) => match t {
                FloatingType::Binary32 => 4,
                FloatingType::Binary64 => 8,
            },
            Type::Pointer { .. } => 8,
            Type::Fn { .. } => 8,
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum PrimitiveType {
    FloatingType(FloatingType),
    IntegralType(IntegralType),
    Bool,
    Void,
}

impl ToString for PrimitiveType {
    fn to_string(&self) -> String {
        match self {
            PrimitiveType::FloatingType(f) => f.to_string(),
            PrimitiveType::IntegralType(i) => i.to_string(),
            PrimitiveType::Bool => "bool".to_owned(),
            PrimitiveType::Void => "void".to_owned(),
        }
    }
}

pub mod tree_visitor {

    use crate::ast::Node;
    use crate::parser::Tree;

    struct Frame {
        node: Node,
        children: Vec<Node>,
    }

    enum PrePost {
        Pre(Node),
        Post(Node),
    }

    /// Don't modify `node` in `pre()`
    /// Don't modify `children` in `pre()`
    pub struct Visitor<'a, F1, F2> {
        tree: &'a mut Tree,
        frames: Vec<Frame>,
        pre: F1,
        post: F2,
    }

    impl<'a, F1, F2> Visitor<'a, F1, F2> {
        pub fn new<T, U>(tree: &'a mut Tree, start: Node, pre: F1, post: F2) -> Visitor<'a, F1, F2>
        where
            F1: FnMut(&mut Tree, Node) -> T,
            F2: FnMut(&mut Tree, Node) -> U,
        {
            let frame = Frame {
                node: Node::MAX,
                children: vec![start],
            };
            Self {
                frames: vec![frame],
                tree,
                pre,
                post,
            }
        }

        fn get_children(&self, node: Node) -> Vec<Node> {
            match self.tree.nodes.get_node(node) {
                super::Tag::FunctionProto {
                    name,
                    parameters,
                    return_type,
                } => {
                    if let Some(params) = parameters {
                        vec![*name, *params, *return_type]
                    } else {
                        vec![*name, *return_type]
                    }
                }
                super::Tag::ParameterList { parameters } => parameters.clone(),
                super::Tag::Parameter { name, ty } => {
                    vec![*name, *ty]
                }
                super::Tag::Pointer { pointee } => {
                    vec![*pointee]
                }
                super::Tag::FunctionDecl { proto, body } => {
                    vec![*proto, *body]
                }
                super::Tag::Block {
                    statements,
                    trailing_expr,
                } => {
                    let mut children = statements.clone();
                    if let Some(expr) = trailing_expr {
                        children.push(*expr);
                    }
                    children
                }
                super::Tag::ReturnStmt { expr } => expr.into_iter().cloned().collect::<Vec<_>>(),
                &super::Tag::ExprStmt { expr } => {
                    vec![expr]
                }
                super::Tag::VarDecl {
                    name,
                    explicit_type,
                    ..
                } => {
                    if let Some(ty) = *explicit_type {
                        vec![*name, ty]
                    } else {
                        vec![*name]
                    }
                }
                super::Tag::GlobalDecl {
                    name,
                    explicit_type,
                    ..
                } => {
                    if let Some(ty) = *explicit_type {
                        vec![*name, ty]
                    } else {
                        vec![*name]
                    }
                }
                &super::Tag::CallExpr { lhs, rhs } => {
                    if let Some(rhs) = rhs {
                        vec![lhs, rhs]
                    } else {
                        vec![lhs]
                    }
                }
                super::Tag::ArgumentList { parameters } => parameters.clone(),
                &super::Tag::Argument { name, expr } => {
                    if let Some(name) = name {
                        vec![name, expr]
                    } else {
                        vec![expr]
                    }
                }
                &super::Tag::ExplicitCast { lhs, typename } => {
                    vec![lhs, typename]
                }
                super::Tag::Deref { lhs }
                | super::Tag::Ref { lhs }
                | super::Tag::Not { lhs }
                | super::Tag::Negate { lhs } => {
                    vec![*lhs]
                }
                super::Tag::Or { lhs, rhs }
                | super::Tag::And { lhs, rhs }
                | super::Tag::BitOr { lhs, rhs }
                | super::Tag::BitAnd { lhs, rhs }
                | super::Tag::BitXOr { lhs, rhs }
                | super::Tag::Eq { lhs, rhs }
                | super::Tag::NEq { lhs, rhs }
                | super::Tag::Lt { lhs, rhs }
                | super::Tag::Gt { lhs, rhs }
                | super::Tag::Le { lhs, rhs }
                | super::Tag::Ge { lhs, rhs }
                | super::Tag::Shl { lhs, rhs }
                | super::Tag::Shr { lhs, rhs }
                | super::Tag::Add { lhs, rhs }
                | super::Tag::Sub { lhs, rhs }
                | super::Tag::Mul { lhs, rhs }
                | super::Tag::Rem { lhs, rhs }
                | super::Tag::Div { lhs, rhs }
                | super::Tag::Assign { lhs, rhs } => {
                    vec![*lhs, *rhs]
                }
                _ => vec![],
            }
        }

        fn next_node(&mut self) -> Option<PrePost> {
            loop {
                let frame = self.frames.last_mut()?;
                if let Some(node) = frame.children.pop() {
                    return Some(PrePost::Pre(node));
                } else {
                    let frame = self.frames.pop()?;
                    if frame.node != Node::MAX {
                        return Some(PrePost::Post(frame.node));
                    }
                }
            }
        }

        pub fn visit_ok<T, E>(mut self) -> core::result::Result<T, E>
        where
            F1: FnMut(&mut Tree, Node) -> core::result::Result<T, E>,
            F2: FnMut(&mut Tree, Node) -> core::result::Result<T, E>,
        {
            let mut t = None;
            loop {
                let Some(node) = self.next_node() else {
                    break;
                };

                match node {
                    PrePost::Pre(node) => {
                        t = Some((self.pre)(self.tree, node)?);
                        let children = self.get_children(node);
                        self.frames.push(Frame { node, children });
                    }
                    PrePost::Post(node) => {
                        t = Some((self.post)(self.tree, node)?);
                    }
                }
            }

            Ok(t.unwrap())
        }

        pub fn visit<T, U>(mut self)
        where
            F1: FnMut(&mut Tree, Node) -> T,
            F2: FnMut(&mut Tree, Node) -> U,
        {
            loop {
                let Some(node) = self.next_node() else {
                    break;
                };

                match node {
                    PrePost::Pre(node) => {
                        (self.pre)(self.tree, node);
                        let children = self.get_children(node);
                        self.frames.push(Frame { node, children });
                    }
                    PrePost::Post(node) => {
                        (self.post)(self.tree, node);
                    }
                }
            }
        }
    }
}
